from ._skiplist import ffi, lib


_undefined = object()


class ObjectRef(object):

    def __init__(self):
        self.__alloc = ffi.new('uintptr_t **')
        self.ref = ffi.cast('void **', self.__alloc)

    @property
    def handle(self):
        return ffi.cast('void *', self.ref[0])

    @property
    def deref(self):
        return ffi.from_handle(self.handle)


class SkipDict(object):

    __slots__ = ('__list', '__pointers')

    def __init__(self):
        self.__list = ffi.new('skiplist_skiplist *')
        self.__pointers = set()
        init_err = lib.skiplist_init(
            self.__list,
            ffi.addressof(lib, 'skiplist_int64_cmp'),
            # Null `void *userdata` for cmp and malloc functions.
            ffi.NULL,
            ffi.NULL)
        if init_err:
            raise RuntimeError("Error initializing skiplist")

    def __del__(self):
        lib.skiplist_free(self.__list)

    def __repr__(self):
        return 'SkipDict({' + ', '.join('{!r}: {!r}'.format(k, v) for k, v in self.iteritems()) + '})'

    def __len__(self):
        return lib.skiplist_size(self.__list)

    def __contains__(self, k):
        return bool(lib.skiplist_find(self.__list, k, ffi.NULL))

    def __nonzero__(self):
        return len(self) > 0

    def __getitem__(self, k):
        result = lib.skiplist_get(self.__list, k, ffi.NULL)
        if result == ffi.NULL:
            raise KeyError(k)
        return ffi.from_handle(result)

    def get(self, k, default=None):
        result = lib.skiplist_get(self.__list, k, ffi.NULL)
        if result == ffi.NULL:
            return default
        return ffi.from_handle(result)

    def __setitem__(self, k, value):
        handle = ffi.new_handle(value)
        self.__pointers.add(handle)
        prev = ObjectRef()
        did_overwrite = lib.skiplist_insert(self.__list, k, handle, prev.ref)
        if did_overwrite:
            self.__pointers.discard(prev.handle)

    def __delitem__(self, k):
        prev = ObjectRef()
        did_remove = lib.skiplist_remove(self.__list, k, prev.ref)
        if not did_remove:
            raise KeyError(k)
        self.__pointers.discard(prev.handle)

    def pop(self, k, default=_undefined):
        prev = ObjectRef()
        did_remove = lib.skiplist_remove(self.__list, k, prev.ref)
        if not did_remove:
            if default is _undefined:
                raise KeyError(k)
            return default
        popped = prev.deref
        self.__pointers.discard(prev.handle)
        return popped

    def __iter__(self):
        for k, v in self.iteritems():
            yield k

    def iteritems(self):
        node = self.__list.head
        while node.next[0]:
            yield (node.next[0].key, ffi.from_handle(node.next[0].val))
            node = node.next[0]

    def minimum(self):
        key_pointer = ffi.new('int64_t *')
        val_ref = ObjectRef()
        has_result = lib.skiplist_min(self.__list, key_pointer, val_ref.ref)
        if not has_result:
            return None
        return (key_pointer[0], val_ref.deref)

    def maximum(self):
        key_pointer = ffi.new('int64_t *')
        val_ref = ObjectRef()
        has_result = lib.skiplist_max(self.__list, key_pointer, val_ref.ref)
        if not has_result:
            return None
        return (key_pointer[0], val_ref.deref)

    def pop_minimum(self):
        key_pointer = ffi.new('int64_t *')
        val_ref = ObjectRef()
        has_result = lib.skiplist_pop(self.__list, key_pointer, val_ref.ref)
        if not has_result:
            return None
        output = (key_pointer[0], val_ref.deref)
        self.__pointers.discard(val_ref.handle)
        return output

    def pop_maximum(self):
        key_pointer = ffi.new('int64_t *')
        val_ref = ObjectRef()
        has_result = lib.skiplist_shift(self.__list, key_pointer, val_ref.ref)
        if not has_result:
            return None
        output = (key_pointer[0], val_ref.deref)
        self.__pointers.discard(val_ref.handle)
        return output



import os

from cffi import FFI


src_root = os.path.join(os.path.dirname(__file__), 'cskipdict', 'headers')
skiplist_header_file = os.path.join(src_root, 'skiplist.cdef')

ffibuilder = FFI()

with open(skiplist_header_file) as header_file:
    ffibuilder.cdef(header_file.read())

ffibuilder.set_source('cskipdict._skiplist', '''
extern "C" {
    #define SKIPLIST_IMPLEMENTATION
    #include "skiplist-config.h"
    #include "skiplist-template.h"
    int skiplist_int64_cmp(int64_t a, int64_t b, void *userdata);

    int skiplist_int64_cmp(int64_t a, int64_t b, void *userdata) {
    return (a > b) - (a < b);
    }
}
''',
include_dirs=[src_root],
sources=[os.path.join(src_root, 'random_uniform.cpp')],
extra_compile_args=['-std=c++11'],
source_extension='.cpp')


if __name__ == '__main__':
    ffibuilder.compile(verbose=True)



from nose.tools import assert_raises

from cskipdict import SkipDict


def check_dict(skip_dict, expected):
    assert len(skip_dict) == len(expected)
    assert len(skip_dict._SkipDict__pointers) == len(set(expected.values()))
    for k in expected:
        assert k in skip_dict
        assert skip_dict[k] == expected[k]
        assert skip_dict.get(k) == expected[k]
        assert skip_dict.get(k + 1000) is None
        assert skip_dict.get(k + 1000, None) is None
        assert skip_dict.get(k + 1000, 'Hello') == 'Hello'
    # Ensure iteration order is always increasing
    last_k = None
    for k in skip_dict:
        assert k in expected
        if last_k is not None:
            assert k > last_k
            last_k = k

    if expected:
        min_key = min(expected.keys())
        assert skip_dict.minimum() == (min_key, expected[min_key])

        max_key = max(expected.keys())
        assert skip_dict.maximum() == (max_key, expected[max_key])
    else:
        assert skip_dict.minimum() is None
        assert skip_dict.maximum() is None


def test_smoke():
    sd = SkipDict()
    check_dict(sd, {})
    sd[123] = 'foo'
    check_dict(sd, {123: 'foo'})
    sd[456] = 'bar'
    check_dict(sd, {123: 'foo', 456: 'bar'})
    sd[123] = 'baz'
    check_dict(sd, {123: 'baz', 456: 'bar'})
    assert sd.pop(456) == 'bar'
    check_dict(sd, {123: 'baz'})
    with assert_raises(KeyError):
        sd.pop(456)
    assert sd.pop(456, None) is None
    assert sd.pop(456, 'Hello') == 'Hello'



from codecs import open
from os import path

from setuptools import setup, find_packages


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    readme = f.read()


setup(
    name='cskipdict',
    version='0.0.4',

    description='A fast CFFI-based SkipDict implementation.',
    long_description=readme,
    url='https://github.com/zacharyvoase/python-cskipdict',

    author='Zachary Voase',
    author_email='zack@meat.io',
    license='UNLICENSE',

    packages=find_packages(exclude=['test']),

    setup_requires=["cffi>=1.4.0"],
    cffi_modules=["cskipdict_build.py:ffibuilder"],

    install_requires=["cffi>=1.4.0"],
)
